/*
** Module   : GlobalSequenceIdentityManager.java
** Abstract : Identity manager implementation which uses a global sequence
**            to generate new IDs
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20080805   @39306 Created initial version. Implementation of
**                           IdentityManager interface which retrieves new
**                           IDs from a global sequence.
** 002 CA  20080815   @39446 Support API change in DBUtils. 
** 003 SVL 20080901   @39647 Added primary key reclamation. Implements
**                           IdentityPoolManager instead of IdentityManager.
** 004 ECF 20090107   @41011 Fixed server shutdown. Timer needed to be a
**                           daemon, it was holding up a clean shutdown of
**                           the server.
** 005 ECF 20090527   @42474 Fixed Context.getStatement(). It was possible to
**                           return a closed statement.
** 006 SVL 20090720   @43238 ProtectedTimer is used instead of Timer.
** 007 SVL 20090930   @44056 If any error has raised into
**                           IdentityPool.nextPrimaryKey() then stop the
**                           idenity pool and use only sequence method for ID
**                           generation.
** 008 SVL 20130331          Upgraded to Hibernate 4.
** 009 ECF 20130510          Fixed sequence access.
*/

package com.goldencode.p2j.persist.id;

import java.math.*;
import java.io.Serializable;
import java.sql.*;
import org.hibernate.*;
import org.hibernate.dialect.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.util.*;
import org.hibernate.internal.*;
import org.hibernate.type.*;

/**
 * Implementation of the {@link IdentityManager} interface which uses a
 * sequence (global to a specific database) to produce 64-bit integer IDs
 * which are unique across all records in that database.
 * <p>
 * Current limitations:
 * <ul>
 *   <li>Wrapping to a negative number after we increment beyond
 *       <code>Long.MAX_VALUE</code> is not handled at this time.
 *   <li>Does not recycle IDs which are allocated but never used.
 *   <li>Does not reclaim IDs for deleted records.
 *   <li>Does not detect/repair fragmentation of the primary key address
 *       space for a table;  existing gaps in primary keys are never plugged.
 *       Note that a solution for this item may preclude having to address the
 *       previous three items.
 * </ul>
 */
public final class SequenceIdentityManager
implements IdentityPoolManager
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(IdentityManager.class);
   
   /** Number of milliseconds before active Hibernate session expires */
   private static long SESSION_TIMEOUT = 5000L; // TODO:  make configurable
   
   /** Object which manages context local sessions and statements */
   private final ContextLocal<Context> context =
   new ContextLocal<Context>()
   {
      protected Context initialValue()
      {
         return (new Context());
      }
   };
   
   /** Timer which executes statement/session timeout tasks */
   private ProtectedTimer timer = null;
   
   /** Identity pool used by this identity manager. */
   private IdentityPool identityPool = null;

   /** Database for which this object generates unique primary keys */
   private Database database = null;
   
   /** Factory which opens new Hibernate sessions */
   private SessionFactory factory = null;
   
   /** SQL statement used to get next sequence value */
   private String nextKeySQL = null;

   /**
    * Default constructor.
    */
   public SequenceIdentityManager()
   {
   }

   /**
    * Specify the database for the identity manager.
    * <code>DefaultIdentityManager</code> is initialized here - current
    * maximum IDs for each table are retrieved.
    *
    * @param   persistence
    *          Persistence service object associated with identity manager.
    */
   public void setPersistence(Persistence persistence)
   {
      Dialect dialect = persistence.getDialect();
      nextKeySQL = dialect.getSequenceNextValString(Persistence.ID_GEN_SEQUENCE);
      database = persistence.getDatabase();
      factory = DatabaseManager.getSessionFactory(database);
      StringBuilder buf = new StringBuilder("SequenceIdentityManager [");
      buf.append(database);
      buf.append("] session closer");
      timer = new ProtectedTimer(buf.toString(), true);
   }

   /**
    * Return the next primary key for the specified table.  The pool of
    * reclaimed IDs is checked first.  If no reclaimed ID is available, a new
    * ID is generated from the sequence.
    *
    * @param   table
    *          Table for which the next primary key will be returned.
    *
    * @return  The next primary key for the given table.
    * 
    * @throws  PersistenceException
    *          if there is an error determining the next primary key ID.
    */
   public Serializable nextPrimaryKey(String table)
   throws PersistenceException
   {
      // Try to get the key from identity pool first.
      if (identityPool != null)
      {
         try
         {
            Serializable key = identityPool.nextPrimaryKey();
            if (key != null)
            {
               return key;
            }
         }
         catch (Throwable thr)
         {
            LOG.error("Identity pool has been stopped because of an error.", thr);
            identityPool.stop();
            identityPool = null;
         }
      }
      
      Context local = context.get();
      
      try
      {
         Long id = null;
         
         synchronized (local)
         {
            PreparedStatement statement = local.getStatement();
            ResultSet rs = statement.executeQuery();
            
            if (!rs.next())
            {
               throw new PersistenceException(
                  "Cannot generate next ID for table '" +
                  database.getName() +
                  "." +
                  table +
                  "'");
            }
            
            id = new Long(rs.getLong(1));
         }
         
         if (LOG.isDebugEnabled())
         {
            Persistence p = PersistenceFactory.getInstance(database);
            LOG.debug(p.message("next primary key:  " + id));
         }
         
         return id;
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         throw new PersistenceException(
            "Error preparing SQL statement:  " + nextKeySQL, exc);
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);
         
         throw new PersistenceException(
            "Error executing SQL statement:  " + nextKeySQL, exc);
      }
   }

   /**
    * Inform the identity manager that the given primary keys can be reused.
    *
    * @param   keys
    *          Keys for reuse.
    */
   public void reclaimKeys(Serializable[] keys)
   {
      if (identityPool != null)
      {
         identityPool.reclaimKeys(keys);
      }
   }

   /**
    * Sets identity pool which can be used by this identity manager.
    *
    * @param   identityPool
    *          Identity pool for use.
    */
   public void setIdentityPool(IdentityPool identityPool)
   {
      Session session = null;
      
      try
      {
         session = factory.openSession();
         
         // Get the next value of sequence
         SQLQuery query = session.createSQLQuery(nextKeySQL);
         ScrollableResults rs = query.scroll();
         
         if (!rs.next())
         {
            throw new HibernateException("Empty result set");
         }
         BigInteger result = (BigInteger) rs.get(0);
         Long sequenceValue = result.longValue();
         
         // Set the next sequence value as a start key for the
         // identity pool and add "wasted" sequence value to the pool.
         identityPool.setScanStartKey(sequenceValue - 1);
         identityPool.reclaimKeys(new Long[] {sequenceValue});
         this.identityPool = identityPool;
      }
      catch (HibernateException exc)
      {
         LOG.error("Cannot get the current sequence value. Indentity pool " +
                   "cannot be started. " + exc);
      }
      finally
      {
         if (session != null)
         {
            session.close();
         }
      }
   }

   /**
    * A class which manages a session and prepared statement which time out
    * and are cleaned up unless used within the timeout period.  This prevents
    * us from having to set up a new session and prepared statement for each
    * request to retrieve the next sequence value, as long as the requests
    * arrive in quick succession.  The timeout prevents us from holding a
    * database connection for too long in the event of intermittent requests.
    */
   private class Context
   {
      /** Current Hibernate session */
      private Session session = null;
      
      /** Statement which retrieves next sequence value */
      private PreparedStatement statement = null;
      
      /** Timestamp of last statement use */
      private long lastUsed = -1L;
      
      /**
       * Get a prepared statement which retrieves the next sequence value.  If
       * no statement currently exists in this context, this method opens a
       * session and prepares a statement and caches both temporarily.  A
       * timer task is added to a timer.  Every SESSION_TIMEOUT milliseconds,
       * the task is executed.  It checks whether the statement has been
       * requested within the timout period.  If not, it closes the cached
       * statement and session, and cancels itself, so that the timer will not
       * execute it again.
       * 
       * @return  Prepared statement which when executed, will return the next
       *          value in the ID generator sequence.
       * 
       * @throws  HibernateException
       *          if there is an error opening a session or getting a JDBC
       *          connection.
       * @throws  SQLException
       *          if there is an error preparing the SQL statement.
       */
      private PreparedStatement getStatement()
      throws SQLException
      {
         if (statement == null)
         {
            assert (timer != null);
            
            session = factory.openSession();
            Connection conn = ((SessionImpl) session).connection();
            statement = conn.prepareStatement(nextKeySQL);
            ProtectedTimerTask task = new ProtectedTimerTask()
            {
               public void run()
               {
                  maybeTimeoutSession(this);
               }
            };
            timer.schedule(task, SESSION_TIMEOUT, SESSION_TIMEOUT);
         }
         
         lastUsed = System.currentTimeMillis();
         
         return statement;
      }
      
      /**
       * This method is invoked on the timer thread every SESSION_TIMEOUT
       * milliseconds after the SQL statement to fetch the next sequence value
       * has been prepared on a new Hibernate session.  If the statement has
       * not been requested during the timeout period, the active statement
       * and session are expired, and the task is cancelled.
       * 
       * @param   task
       *          ProtectedTimerTask to be cancelled upon expiry of the
       *          session and prepared statement.
       */
      private synchronized void maybeTimeoutSession(ProtectedTimerTask task)
      {
         long now = System.currentTimeMillis();
         if (now - lastUsed >= SESSION_TIMEOUT)
         {
            // The statement timeout period has elapsed.  Close the statement
            // and kill the timer.
            try
            {
               if (LOG.isDebugEnabled())
               {
                  LOG.debug(
                     "Closing primary key generator session after timeout");
               }
               if (statement != null)
               {
                  statement.close();
                  statement = null;
               }
               if (session != null)
               {
                  session.close();
                  session = null;
               }
            }
            catch (HibernateException exc)
            {
               DBUtils.handleException(database, exc);
            }
            catch (SQLException exc)
            {
               DBUtils.handleException(database, exc);
            }
            finally
            {
               task.cancel();
            }
         }
      }
   }
}
