/*
** Module   : ImportWorker.java
** Abstract : Pattern worker used to import Progress data into a relational
**            database using Hibernate batch insert
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050715   @21822 Created initial version. Includes support to
**                           insert new records in batch mode, and to
**                           query individual records in existing tables
**                           to resolve foreign associations.
** 002 ECF 20050801   @21970 Substantial rewrite to add fault tolerance
**                           and improve performance. Much of the function
**                           previously implemented in the import.xml
**                           ruleset has been moved into the importTable
**                           user function.
** 003 ECF 20050805   @22014 Minor changes to error logging and progress
**                           indicator. Also some doc additions.
** 004 ECF 20050809   @22078 Added analyze step to end of table import.
**                           Runs the postgres-specific "analyze" command.
**                           This eventually will need to be made vendor
**                           neutral.
** 005 ECF 20050810   @22081 Modified QueryHelper and calling code to fix
**                           parameter name mismatch. In some cases, pairs
**                           of properties involved in a join of two
**                           classes can have their names fall out of
**                           synch during name conversion. Modified the
**                           parameter management in QueryHelper to handle
**                           this disconnect. Also introduced a separate
**                           session for foreign association queries only,
**                           which is never flushed and cleared much less
**                           often. This should help query performance,
**                           as objects can be retrieved from the first
**                           level cache more often.
** 006 ECF 20050819   @22205 Removed RecordLoader and PropertyMapper inner
**                           classes (and all subclasses). These classes
**                           were abstracted into their own top level
**                           classes to support sharing between database
**                           and directory import tasks. Created inner
**                           class SqlRecordLoader which extends the new
**                           RecordLoader class to handle database import.
** 007 ECF 20050930   @22962 Added dynamic creation of indexes during
**                           table import. This was necessary to work
**                           around a Hibernate limitation which does not
**                           guarantee the order of index columns during
**                           normal schema DDL generation. Also, by doing
**                           it this way, we avoid the performance hit of
**                           updating all non-unique indexes for each
**                           insert. Instead, we build only unique indexes
**                           up front (to ensure data integrity), then
**                           defer creation of the remaining indexes until
**                           after the data has been imported. Also added
**                           use of Hibernate's query cache to speed up
**                           queries against tables which have one-to-many
**                           associations with tables being imported.
** 008 ECF 20051027   @23201 Added support for case-insensitive index
**                           components. Use the database upper() function
**                           to ignore case when creating indexes on case-
**                           insensitive text columns.
** 009 ECF 20051103   @23216 Removed user functions. nonRedundantIndexes
**                           and normalizeUniqueIndexes were removed.
** 010 ECF 20060713   @28039 Made improvements in indexing support. Added
**                           reindexing feature, which drops and recreates
**                           all indexes defined for each database table.
** 011 ECF 20060714   @28137 Updated QueryHelper inner class. Changes were
**                           necessary to emit HQL using the new RTRIM
**                           syntax embedded for character properties.
** 012 ECF 20070312   @32389 Multi-threaded the import worker. Each table
**                           is imported in a dedicated thread. The number
**                           of threads which may run simultaneously is
**                           configurable. This change significantly
**                           improves the throughput of the import job on
**                           multi-processor hardware. There should be
**                           some improvement on a uni-processor system as
**                           well, though it will not be as dramatic.
** 013 ECF 20070314   @32411 Reworked import job prioritization. Jobs are
**                           now ranked and sorted to ensure that critical
**                           import paths are optimized for maximum
**                           throughput. Also reworked job dispatching
**                           model to proactively schedule jobs, rather
**                           than leaving this up to the JVM's thread
**                           scheduler.
** 014 ECF 20070424   @33214 Removed text-only optimization in reindex().
**                           In some cases, re-establishing all indices
**                           for a table is necessary.
** 015 EVL 20070609   @34005 Adding explicit import of the class
**                           org.hibernate.type.Type to eliminate conflict
**                           with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 016 ECF 20070705   @34362 Minor optimization. Replaced StringBuffer
**                           with StringBuilder.
** 017 ECF 20070908   @35340 Fixed memory problem. Use SymbolResolver
**                           constructor which disables SchemaDictionary.
**                           We don't need schema lookups and that class
**                           is not thread-safe, so we were creating a
**                           huge memory footprint by creating multiple
**                           SchemaDictionary instances across threads.
**                           Replaced record counter ints with longs.
** 018 ECF 20080104   @36679 Set ErrorManager to headless mode. Better
**                           error reporting for errors encountered while
**                           loading next record from export data.
** 019 ECF 20080306   @37467 Modified to support changes in DBUtils.
** 020 GES 20080418   @38172 Moved from a lexer approach to using the
**                           equivalent runtime support for the Progress
**                           IMPORT language stmt. This allows the import
**                           process to match the behavior of Progress
**                           which seems to internally use the IMPORT
**                           stmt. Note that this allows proper date
**                           component ordering, non-ASCII character set
**                           and other I18N support to work properly.
** 021 ECF 20080804   @39292 Reworked primary key generation strategy. All
**                           primary keys generated for a given database
**                           must be unique across all tables.
** 022 ECF 20081124   @40686 Remove analyze command after table import.
**                           This is no longer necessary, since we are not
**                           immediately running queries for foreign key
**                           lookups against newly imported tables. On
**                           older versions of PostgreSQL, this was
**                           causing conflicts with autovacuum.
** 023 GES 20090422   @41898 Converted to standard string formatting.
** 024 GES 20100630          Refactored code to expose a new method called
**                           generateIndexDDL(). This allows the DDL which is
**                           normally created during data import to be captured
**                           to an output stream. Cleaned up some unnecessary
**                           usage of database metadata during initialize().
** 025 CA  20101026          Moved the CREATE/DROP INDEX statements to be 
**                           generated by the dialect. Set the default value 
**                           for any not-null property which has missing data.
** 026 OM  20121210          Added import sequence value.
** 027 ECF 20121212          Moved setting of date-order from runImport to
**                           importAsync. This setting is context-local, so it
**                           has to be set on the job thread. Skip word index
**                           in {create|drop}Indexes methods. Fixed import of
**                           sequence values to consider name conversion.
**                           Fixed import of lone hyphen entry from an export
**                           file (was being skipped).
** 028 SVL 20130331          Upgraded to Hibernate 4.
** 029 ECF 20130510          Fixed sequence initialization, which would error out for some
**                           database dialects.
*/

package com.goldencode.p2j.schema;

import java.io.*;
import java.lang.InstantiationException;
import java.lang.reflect.*;
import java.sql.*;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.*;
import org.hibernate.cfg.*;
import org.hibernate.dialect.*;
import org.hibernate.service.*;
import org.hibernate.type.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.type.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;

/**
 * A pattern worker that provides batch data import services. Exposes user
 * functions which assist in inserting data read from Progress
 * <code>*.d</code> data dump files into a relational database using
 * Hibernate.  Defines a number of inner, helper classes, most of which are
 * configured before a table is imported.
 * <p>
 * This worker enables a multi-threaded import, where different tables are
 * loaded in parallel threads.  The number of threads which may actively be
 * importing data at once is set with the {@link
 * ImportWorker.Library#setMaximumThreads setMaximumThreads} method, which is
 * exposed to TRPL as a user function.
 * <p>
 * The idiom used to run an import job is to walk through the P2O tree of the
 * target database using the <code>schema/import.xml</code> ruleset.  This
 * ruleset prepares the necessary {@link ImportWorker.SqlRecordLoader}
 * instances and lists of indexes which must be created for each table.  It
 * then invokes the {@link ImportWorker.Library#prepareImport prepareImport}
 * user function for each table.  This creates the necessary data structures
 * in preparation for the import.
 * <p>
 * When all tables have been prepared, the import ruleset invokes the {@link
 * ImportWorker.Library#runImport runImport} user function.  This analyzes
 * the tables targeted for import, their associated data export files, and
 * their dependencies and dependent tables to determine the critical paths
 * which must be traversed during the import.  The tables are then sorted
 * into priority order to ensure that all critical paths are traversed as
 * directly as possible.
 * <p>
 * Once tables have been prioritized and sorted, the import begins.  The
 * prioritized queue of import jobs is traversed in order, and a thread is
 * launched for each import job in the queue, up to the maximum number of
 * active threads.  When a thread completes importing a table, the next
 * available job is removed from the queue and processed.  A job is only
 * allowed to commence if all tables upon which it is dependent have been
 * imported first.  If this is not the case, this job is skipped and the next
 * waiting job is tested.  This continues until a job is found which can
 * commence immediately, or until the end of the job queue is reached.  If
 * no available job was found, but the job queue is not empty, the main
 * thread waits until one of the active threads completes its job, then runs
 * through the queue again, looking for an available candidate.  This process
 * continues until the job queue is empty.
 * <p>
 * Once all active import threads have completed and no more jobs are waiting
 * in the job queue, the import run is complete and the <code>runImport</code>
 * method returns control to TRPL.
 * <p>
 * Progress statistics are maintained and reported at the completion of each
 * job in the queue.  These include percent complete (based upon export file
 * bytes processed), records loaded, and average number of records processed
 * across various periods.  Because these are only calculated and reported
 * upon completion of a job, these statistics trend lower than the actual
 * throughput being achieved.  This is because at the point where one thread
 * completes its work for a single table, all other threads are at some state
 * of partial completion of their own jobs, so the actual throughput is not
 * accurately reflected in the current calculations.  However, the trade-off
 * for more accurate, real-time statistics would have been a higher level of
 * synchronization and a more complicated algorithm to account for dropped
 * records and actual records imported.  Rather than possibly compromise
 * throughput with such an implementation, it was deemed better to settle for
 * less accurate statistics.  The final set of statistics reported at the end
 * of the full run is reflective of the actual, average performance achieved.
 * <p>
 * Moving from a single-threaded (as this worker originally was designed) to
 * a multi-threaded architecture paid huge dividends, particularly on
 * multi-processor hardware.  On a 4-processor, dual-core system, running
 * with 8 active threads, an import of approximately 79 million primary
 * records completed in approximately 40% of the time needed for a
 * single-threaded run on the same system.  Gains on a uni-processor system
 * are not expected to be as dramatic, though some slight improvement is
 * likely.  Performance tends to scale directly with the number of (virtual)
 * processors on a system.  It is recommended to run with one thread per
 * virtual processor for best results.
 */
public final class ImportWorker
extends AbstractPatternWorker
{
   /** Logger */
   private static final Log log = LogFactory.getLog(ImportWorker.class);
   
   /** Default JDBC batch size if not in Hibernate configuration */
   private static final int DEFAULT_BATCH_SIZE = 50;
   
   /** Database metadata table type for regular tables */
   private static final String TABLE = "TABLE";
   
   /** Database metadata column name for table name metadata column */
   private static final String TABLE_NAME = "TABLE_NAME";
   
   /** Database metadata column name for column name metadata column */
   private static final String COLUMN_NAME = "COLUMN_NAME";
   
   /** Database metadata column name for data type metadata column */
   private static final String DATA_TYPE = "DATA_TYPE";
   
   /** Format specifier for an import statistics update message. */
   private static final String STATS_FMT =
      "%.5f%% complete:  %d recs in %d:%02d:%06.3f (%d/sec, %d/min, %d/hr)";
   
   /** Map of unqualified DMO names to import bundles */
   private final Map<String, ImportBundle> importBundles =
      new LinkedHashMap<String, ImportBundle>();

   /** Map of historical to SQL sequence names */
   private final Map<String, String> sequences =
      new HashMap<String, String>();

   /** Set of unqualified DMO class names for tables which were processed */
   private final Set<String> complete = new HashSet<String>();
   
   /** Set of unqualified DMO class names for tables which had fatal errors */
   private final Set<String> fatal = new HashSet<String>();
   
   /** Set of all DMO classes for which an import was attempted */
   private final Set<String> attempted = new HashSet<String>();
   
   /** Lock for record count synchronization */
   private final Object countLock = new Object();
   
   /** Deque of IDContexts, up to maximumThreads deep */
   private Deque<IDContext> idContextStack =
      new ArrayDeque<IDContext>();
   
   /** Queue of import jobs */
   private List<ImportBundle> jobQueue = null;
   
   /** Hibernate session factory */
   private SessionFactory sessionFactory = null;
   
   /** JDBC batch size */
   private int batchSize = 0;
   
   /** Progress indicator increment */
   //private int increment = 500000;
   
   /** Database dialect in use */
   private P2JDialect dialect = null;
   
   /** Number of import threads allowed to be active simultaneously */
   private int maximumThreads = 4;
   
   /** Number of import threads currently active */
   private int activeThreads = 0;
   
   /** Number of records imported */
   private long imported = 0;
   
   /** Number of bytes imported */
   private double importedBytes = 0.0;
   
   /** Total byte count of all exported data files */
   private double totalBytes = 0.0;
   
   /** System time at beginning of import */
   private long startTime = 0L;
   
   /** Next ID bracket */
   private long nextGlobalID = 0L;
   
   /** ID bracket size */
   private long idBracketSize = 10000L;
   
   /**
    * Default constructor which initializes callback library.
    */
   public ImportWorker()
   {
      super();
      
      setLibrary(new Library());
      ErrorManager.setHeadless();
   }
   
   /**
    * Open a new database session and set it to flush only on explicit
    * flushes.
    *
    * @return  New session.
    */
   private Session openSession()
   {
      Session session = sessionFactory.openSession();
//      session.setFlushMode(FlushMode.MANUAL);
      session.setFlushMode(FlushMode.NEVER);
      
      return session;
   }
   
   /**
    * Build the dialect-specific DDL to drop the primary ID sequence.
    *
    * @param    dialect
    *           The database-specific SQL helper.
    *
    * @return   The generated DDL.
    */
   private String buildDropIdentitySequence(Dialect dialect)
   {
      String[] ds = dialect.getDropSequenceStrings(Persistence.ID_GEN_SEQUENCE);
      return ds[0];
   }
   
   /**
    * Build the dialect-specific DDL to create the primary ID sequence.
    *
    * @param    dialect
    *           The database-specific SQL helper.
    * @param    highID
    *           The starting value for the sequence.
    *
    * @return   The generated DDL.
    */
   private String buildCreateIdentitySequence(P2JDialect dialect, long highID)
   {
      return dialect.getCreateSequenceString(Persistence.ID_GEN_SEQUENCE,
                                             highID);
   }
   
   /**
    * Build the DDL to create the primary ID sequence and write this SQL to the
    * given stream.
    *
    * @param    os
    *           Output destination.
    * @param    delimiter
    *           End of statement delimiter.
    * @param    dialect
    *           The database-specific SQL helper.
    * @param    highID
    *           The starting value for the sequence.
    */
   private void generateIdentitySequence(OutputStream os,
                                         String       delimiter,
                                         P2JDialect   dialect,
                                         long         highID)
   throws java.io.IOException,
          java.io.UnsupportedEncodingException
   {
      String dropSeq   = buildDropIdentitySequence((Dialect) dialect);
      String createSeq = buildCreateIdentitySequence(dialect, highID);
      writeStmt(os, dropSeq, delimiter);
      writeStmt(os, createSeq, delimiter);
   }
   
   /**
    * Determines the high-water primary key generated during the import and
    * creates an identity generator sequence whose starting value is that ID.
    * Presumes backing database supports sequences.
    */
   private void createIdentitySequence()
   {
      long highID = 1L;
      for (IDContext idCtx : idContextStack)
      {
         highID = Math.max(highID, idCtx.getCurrentID());
      }
      
      String ddl = buildCreateIdentitySequence((P2JDialect) dialect, highID);
      
      if (ddl == null)
      {
         System.err.println(
            "Creation of identity generator sequence starting at " +
            highID +
            " is unsupported");
         
         return;
      }

      Session session = openSession();
      Transaction tr = null;

      try
      {
         tr = session.beginTransaction();
         session.createSQLQuery(ddl).executeUpdate();
         tr.commit();
      }
      catch (HibernateException exc)
      {
         if (tr != null)
            tr.rollback();
         exc.printStackTrace();
      }
      finally
      {
         if (session != null && session.isOpen())
         {
            session.close();
         }
      }
   }
   
      
   /**
    * Open a new buffered output stream for the given target filename.
    *
    * @param    filename
    *           Target output file.  Must not reference a directory or any
    *           resource that cannot be created or written to.
    *
    * @return   The newly opened stream.
    *
    * @throws   IllegalArgumentException
    *           If the file cannot be properly accessed or opened.
    */
   private OutputStream openStream(String filename)
   {
      File         file = new File(filename);
      OutputStream out  = null;
      
      if (file.isDirectory())
      {   
         throw new IllegalArgumentException(filename + " is a directory!");
      }
      
      try
      {
         // this is a file, get the containing directory
         File parent = file.getParentFile();
         
         if (parent != null && !parent.exists())
         {
            // the directory doesn't exist, so create it (including all
            // intermediate directories needed
            parent.mkdirs();
         }
         
         // create/overwrite the file
         FileOutputStream fos = new FileOutputStream(file, true);
         
         // init our buffered stream wrapper 
         out = new BufferedOutputStream(fos);
      }
      
      catch (Exception exc)
      {
         throw new IllegalArgumentException("Cannot create stream for " +
                                            filename + "(" +
                                            exc.toString() + ")!", exc);
      }
      
      return out;
   }
   
   /**
    * Output the given text to the stream in UTF-8 encoding.
    *
    * @param    os
    *           Output destination.
    * @param    text
    *           Content to write.
    */
   private void write(OutputStream os, String text)
   throws java.io.IOException,
          java.io.UnsupportedEncodingException
   {
      os.write(text.getBytes("UTF-8"));
   }
   
   /**
    * Output the given text as a statement (delimited and with a new line) to
    * the stream in UTF-8 encoding.
    *
    * @param    os
    *           Output destination.
    * @param    text
    *           Content to write.
    * @param    delimiter
    *           Statement ending delimiter.
    */
   private void writeStmt(OutputStream os, String text, String delimiter)
   throws java.io.IOException,
          java.io.UnsupportedEncodingException
   {
      StringBuilder sb = new StringBuilder();
      sb.append(text).append(delimiter).append("\n");
      write(os, sb.toString());
   }
   
   /**
    * Library of callback user functions exposed to rulesets.
    */
   public class Library
   {
      /**
       * Configure the given Hibernate configuration object using the
       * information specified in the given configuration file and build
       * the resulting Hibernate session factory. One particular parameter,
       * JDBC batch size, is explicitly configured to a default value if it
       * is not specified in the configuration file.
       * <p>
       * In addition, the names of all text columns are mapped by table name
       * for use during index creation.
       *
       * @param   config
       *          Hibernate configuration object.
       * @param   props
       *          The Hibernate properties.
       *
       * @return  Hibernate session factory used to create sessions for
       *          database access.
       *
       * @throws  SQLException
       *          if any error occurs collecting database metadata.
       * @throws  HibernateException
       *          if any error occurs configuring Hibernate or building the
       *          session factory.
       */
      public SessionFactory initialize(Configuration config, Properties props)
      throws SQLException
      {
         config.addProperties(props);
         ServiceRegistry registry = new ServiceRegistryBuilder()
				                        .applySettings(props)
				                        .buildServiceRegistry();
         Settings settings = config.buildSettings(registry);

         // Save off dialect for later use.
         dialect = (P2JDialect) Dialect.getDialect(props);
         
         // Ensure batch size is set to a reasonable value for import.
         batchSize = settings.getJdbcBatchSize();
         if (batchSize == 0)
         {
            batchSize = DEFAULT_BATCH_SIZE;
            config.setProperty("hibernate.jdbc.batch_size",
                               String.valueOf(batchSize));
            log.info("Using default batch size for import");
         }
         
         //increment = Math.max(increment, batchSize);
         sessionFactory = config.buildSessionFactory(registry);
         
         startTime = System.currentTimeMillis();
         
         return sessionFactory;
      }
      
      /**
       * Set the maximum allowable threads.  This sets the number of threads
       * which may run table imports simultaneously.  However, a larger
       * number of threads may be created.  Additional threads may block
       * waiting for table upon which they are dependent to complete their
       * respective imports.
       *
       * @param   maximumThreads
       *          Maximum number of import threads which may be active at one
       *          time.
       */
      public void setMaximumThreads(int maximumThreads)
      {
         synchronized (complete)
         {
            ImportWorker.this.maximumThreads = maximumThreads;
         }
      }
      
      /**
       * Store the given objects in preparation for an import run.  The
       * import for all prepared tables commences, potentially on multiple
       * threads, when the {@link #runImport} method is invoked.
       *
       * @param   loader
       *          Record loader helper which contains all information
       *          necessary to find the import data and to map each record
       *          in the import file to a DMO of the appropriate class.
       * @param   indexes
       *          A list of index definitions which must be applied to the
       *          table before (for unique indexes) and after (for non-unique
       *          indexes) data import.
       *
       * @see     #setMaximumThreads
       */
      public void prepareImport(SqlRecordLoader loader, List indexes)
      {
         List<P2JIndex> idxList = (List<P2JIndex>) indexes;
         importBundles.put(loader.getDMOName(),
                           new ImportBundle(loader, idxList));
      }
      
      /**
       * Commence an import run, potentially on multiple threads.  All tables
       * prepared by {@link #prepareImport} will be imported.
       * <p>
       * This method organizes the tables into an order which is intended to
       * maximize throughput throughout the import run, by scheduling those
       * tables with the most time-consuming import path to run earliest in
       * the process.
       * <p>
       * Multiple, parallel threads are used to import multiple tables at
       * once.
       * <p>
       * This method will block until all prepared tables have been imported.
       *
       * @return  The number of records successfully imported.
       *
       * @throws  InterruptedException
       *          if the thread on which this method is invoked is
       *          interrupted.
       *
       * @see     #prepareImport
       * @see     #setMaximumThreads
       */
      public long runImport()
      throws InterruptedException
      {
         // Setup JVM-wide configuration values to support certain runtime
         // behavior that would otherwise come from the directory.
         
         String dorder = com.goldencode.p2j.cfg.Configuration
                            .getParameter("date-order");
         
         String gsep = com.goldencode.p2j.cfg.Configuration
                          .getParameter("number-group-sep");
         String dsep = com.goldencode.p2j.cfg.Configuration
                          .getParameter("number-decimal-sep");
         
         if (gsep != null && gsep.length() == 1 &&
             dsep != null && dsep.length() == 1)
         {
            NumberType.overrideDefaultSeparators(gsep.charAt(0),
                                                 dsep.charAt(0));
         }
         
         // This iteration is guaranteed to visit bundles which are
         // dependencies of other tables, before visiting their dependent
         // bundles.
         Iterator<ImportBundle> iter = importBundles.values().iterator();
         while (iter.hasNext())
         {
            ImportBundle bundle = iter.next();
            totalBytes += bundle.getFileLength();
            bundle.calculateWeight();
            Set<String> dependencies = bundle.getLoader().getDependencies();
            Iterator<String> iter2 = dependencies.iterator();
            while (iter2.hasNext())
            {
               String dep = iter2.next();
               ImportBundle depBundle = importBundles.get(dep);
               if (depBundle != null)
               {
                  depBundle.addDependent(bundle);
               }
            }
         }
         
         List<ImportBundle> sorted =
            new ArrayList<ImportBundle>(importBundles.values());
         
         // Calculate relative ranks for each import bundle.  This will
         // determine the order in which the tables are loaded.
         iter = sorted.iterator();
         while (iter.hasNext())
         {
            ImportBundle next = iter.next();
            next.calculateRank();
         }
         
         // Sort in priority order, highest priority first.
         Collections.sort(sorted);
         
         // Report the order in which the jobs will be performed.
         System.out.println("IMPORT ORDER:");
         iter = sorted.iterator();
         for (int i = 1; iter.hasNext(); i++)
         {
            ImportBundle next = iter.next();
            System.out.println(String.valueOf(i) + ") " + next);
         }
         
         // Set up the job queue.
         jobQueue = new LinkedList<ImportBundle>(sorted);
         ImportBundle job = null;
         
         // Schedule asynchronous imports of the import jobs, in order.
         while ((job = nextJob()) != null)
         {
            importAsync(job, dorder);
         }
         
         // Wait for all jobs to finish, reporting any fatal errors which
         // were encountered.  Create a sequence for identity generation,
         // using the high-water mark ID created during the import.
         synchronized (complete)
         {
            while (!complete.containsAll(attempted))
            {
               complete.wait();
            }
            
            if (!fatal.isEmpty())
            {
               System.out.println("The following tables had fatal errors:  " +
                                  fatal);
            }
            else
            {
               createIdentitySequence();
            }
         }
         
         // Return the number of records imported across all jobs.
         synchronized (countLock)
         {
            return imported;
         }
      }

      /**
       * Adds a sequence to the map of sequences to be initialized from the
       * sequences export file.
       *
       * @param   historical
       *          the original name of the sequence.
       * @param   sqlName
       *          the converted, SQL-compliant name of the sequence.
       */
      public void addSequence(String historical, String sqlName)
      {
         sequences.put(historical, sqlName);
      }

      /**
       * Sets the values of the  sequences. All sequences must be already
       * added to the list.
       *
       * @param   seqDumpFile
       *          The file to which the sequences' values have been exported
       *
       * @return  The number of successfully initialized
       */
      public long initializeSequences(String path, String seqDumpFile)
      {
         long count = 0;
         
         Scanner scanner;
         try
         {
            scanner = new Scanner(new FileInputStream(path + seqDumpFile));
         }
         catch (FileNotFoundException e)
         {
            return 0;
         }
         
         Session session = null;
         try
         {
            while (scanner.hasNextLine())
            {
               String line = scanner.nextLine();
               if (line.equals("."))
               {
                  break;
               }
               
               String[] tokens = line.split("\\s");
               if (tokens.length != 3)
               {
                  // malformed line, must have exactly 3 tokens
                  System.err.println(
                     "Invalid line in '" +
                     seqDumpFile +
                     "'. The file may not be a sequence export file.");
                  
                  continue;
               }
               
               // interpreting the 3 tokens
               int seqId = Integer.parseInt(tokens[0]);
               
               // remove both " from begin and end
               String histName = tokens[1].substring(1, tokens[1].length() - 1);
               long seqVal = Long.parseLong(tokens[2]);
               
               if (!sequences.containsKey(histName))
               {
                  // that's strange, the two lists do not match
                  System.err.println(
                     "Unknown properties for sequence '" +
                     histName + 
                     "'. The sequence does not appear in .df export so " +
                     "the value from '" +
                     seqDumpFile +
                     "' will be ignored.");
                  
                  continue;
               }
               
               String sqlName = sequences.get(histName);
               String query = dialect.getSequenceSetValString(sqlName, seqVal);
               if (session == null)
               {
                  // lazy initialization
                  session = openSession();
               }
               
               Transaction tr = null;
               try
               {
                  System.out.println("Importing sequence '" + sqlName + "'...");
                  tr = session.beginTransaction();
                  SQLQuery sqlQ = session.createSQLQuery(query);
                  if (query.trim().toUpperCase().startsWith("SELECT"))
                  {
                     // a select statement for this likely means we are calling a function to set
                     // the value of the sequence;  we don't care about the result, just the
                     // side-effect of executing the function
                     sqlQ.list();
                  }
                  else
                  {
                     sqlQ.executeUpdate();
                  }
                  tr.commit();
                  
                  count++; // increment counter on success
                  sequences.remove(histName);
               }
               catch (HibernateException exc)
               {
                  if (tr != null)
                     tr.rollback();
                  exc.printStackTrace();
               }
            }
         }
         finally
         {
            scanner.close();
            if (session != null && session.isOpen())
            {
               session.close();
            }
         }
         
         // double check if all sequence have been initialized
         if (!sequences.isEmpty())
         {
            System.err.println(
               "The following sequences were NOT initialized because they " +
               "were not found in the '" +
               seqDumpFile + 
               "' export file or import failed:");
            for (String seq : sequences.keySet())
            {
               System.err.println(seq);
            }
         }
         
         return count;
      }
      
      /**
       * Import all data for a table into the database, using the given record
       * loader and current session factory. Uses the {@link Stream#readField}
       * to scan the input file. Each record's data is stored in a new
       * instance of a data model object (DMO). Hibernate is used to persist
       * the DMO to the backing database.
       * <p>
       * If the DMO refers to any other DMOs, the necessary queries are
       * executed to resolve these references before the object is
       * persisted. This ensures foreign key associations are correctly
       * resolved at import time.
       * <p>
       * The import uses JDBC batching via Hibernate to send data to the
       * database in batches rather than a single record at a time. Batch
       * size is retrieved from the Hibernate configuration, and is set to
       * a default if not specified in the config.
       * <p>
       * This method is designed to be fairly fault tolerant. Database errors
       * which are the result of data which does not conform to the expected
       * schema generally are handled by dropping the offending record,
       * adding a log entry to report the nature of the error, and continuing
       * with the next record, if any. Specifically, the following conditions
       * are recoverable:
       * <ul>
       * <li>insert which violates an explicit unique constraint
       * <li>insert which violates an explicit not-null constraint
       * <li>query which is supposed to return a unique result returns
       *     more than one result
       * </ul>
       * <p>
       * The following errors are not recoverable:
       * <ul>
       * <li>data which cannot be lexed
       * <li>data which does not match the expected lexical token type
       * <li>any other exceptions thrown by Hibernate or the database
       * </ul>
       * In the event an unrecoverable error occurs, the records in the
       * current batch and all the remaining records for the current table
       * are not imported. However, all records imported up to that point
       * are preserved in the database.
       * <p>
       * Errors are logged as configured for the current VM.
       *
       * @param   loader
       *          Record loader helper which contains all information
       *          necessary to find the import data and to map each record
       *          in the import file to a DMO of the appropriate class.
       * @param   indexes
       *          A list of index definitions which must be applied to the
       *          table before (for unique indexes) and after (for non-unique
       *          indexes) data import.
       *
       * @return  The number of records successfully imported.
       */
      public int importTable(SqlRecordLoader loader, List<P2JIndex> indexes)
      {
         IDContext idCtx;
         if (idContextStack.peek() != null)
         {
            idCtx = idContextStack.pop();
         }
         else
         {
            idCtx = new IDContext();
         }
         Session session = null;
         Stream stream = null;
         int counter = 0;
         List<Object> records = new ArrayList<Object>(batchSize);
         File dataFile = loader.getDataFile();
         int badIndex = -1;
         int dropCount = 0;
         long nextID = idCtx.getCurrentID();
         try
         {
            // Attempt to drop indexes.
            List<P2JIndex> nonRedundantIndexes =
               P2JIndex.nonRedundantIndexes(indexes);
            dropIndexes(nonRedundantIndexes);
            
            // Attempt to create unique indexes before data import.
            session = openSession();
            List<P2JIndex> minimalUnique =
               P2JIndex.normalizeUniqueIndexes(indexes);
            createIndexes(session, minimalUnique, true);
            Map<String, BaseDataType> notNullProperties = loader.notNullProps;
            boolean hasNotNull = notNullProperties.size() != 0;

            // Process data records.
            stream =
               new FileStream(dataFile.getPath(), false, false)
               {
                  // don't skip a lone hyphen read from an export file
                  protected boolean skipLoneHyphenInput() { return false; }
               };
            
            boolean eof = false;
            while (true)
            {
               // Reset bad index to clear error status;  will be set to a
               // non-negative number on error.
               badIndex = -1;
               
               // Open a new session if (a) this is the first time through,
               // or (b), the old one was closed due to an exception.
               if (session == null)
               {
                  session = openSession();
               }
               Transaction tx = session.beginTransaction();
               
               // Read up to batchSize records from input file.
               for (int i = records.size();
                    i < batchSize && !eof;
                    i++, counter++)
               {
                  // Read the next record.
                  Object dmo = null;
                  try
                  {
                     dmo = loader.nextRecord(stream, session);
                  }
                  catch (NonUniqueResultException exc)
                  {
                     log.error(
                        "Dropped record #" + (counter + 1) + " in " +
                        dataFile.getName() + " due to error:  " +
                        exc.getMessage());
                     
                     badIndex = counter;
                     dropCount++;
                     
                     // This error is recoverable;  no need to discard
                     // session, just continue.
                     continue;
                  }
                  catch (Exception exc)
                  {
                     if (session.isOpen())
                     {
                        session.close();
                     }
                     session = null;
                     
                     // Not recoverable;  log and abort.
                     log.error(
                        "Error creating record " + nextID +
                        " (" + dropCount + " previously dropped) [" +
                        stream.getPos() + " byte index ]");
                     
                     throw exc;
                  }
                  
                  if (dmo == null)
                  {
                     eof = true;
                     break;
                  }
                  
                  if (hasNotNull)
                  {
                     // for all not-null properties which don't have data, set
                     // it to default
                     for (String prop : notNullProperties.keySet())
                     {
                        PropertyMapper pm = loader.getProperty(prop);
                        BaseDataType bdt = pm.invokeGetter(dmo);
                        if (bdt == null || bdt.isUnknown())
                        {
                           Object def = notNullProperties.get(prop);
                           if (def != null)
                           {
                              pm.invokeSetter(dmo, def);
                           }
                        }
                     }
                  }
                  
                  records.add(dmo);
               }
               
               // Abort if nothing to do.
               if (records.size() == 0)
               {
                  break;
               }
               
               // Import the batch as a discrete transaction.
               int index = 0;
               try
               {
                  // Associate each transient object with the session.
                  for (int size = records.size(); index < size; index++)
                  {
                     Persistable dmo = (Persistable) records.get(index);
                     nextID = idCtx.getNextID();
                     dmo.setId(nextID);
                     session.save(dmo);
                  }
                  
                  // Flush batch of inserts and release memory.
                  session.flush();
                  session.clear();
                  
                  // Commit transaction.
                  tx.commit();
                  
                  /*
                  // Primitive progress indicator.
                  if ((counter - dropCount) % (increment) == 0)
                  {
                     System.out.print(".");
                  }
                  */
                  
                  // Clear record list for next pass.
                  records.clear();
               }
               catch (PropertyValueException exc)
               {
                  tx.rollback();
                  
                  records.remove(index);
                  dropCount++;
                  badIndex = index;
                  
                  log.error(
                     "Dropped record #" +
                     (counter - Math.min(counter, batchSize) + dropCount +
                      index) + " in " + dataFile.getName() +
                     " due to error:  " + exc.getMessage());
                  
                  // Close and discard the failed session.
                  if (session.isOpen())
                  {
                     session.close();
                  }
                  session = null;
               }
               catch (HibernateException exc)
               {
                  badIndex = 0;
                  tx.rollback();
                  badIndex = removeFailingRecord(records,
                                                 exc,
                                                 counter,
                                                 dataFile);
                  dropCount++;
                  
                  // Close and discard the failed session.
                  if (session.isOpen())
                  {
                     session.close();
                  }
                  session = null;
               }
            }
            
            if (session == null)
            {
               session = openSession();
            }
            
            // Attempt to create remaining indexes.
            Set<P2JIndex> remainingIndexes =
               new LinkedHashSet<P2JIndex>(nonRedundantIndexes);
            remainingIndexes.removeAll(minimalUnique);
            createIndexes(session, remainingIndexes, false);
            
            // Analyze the table to speed queries which leverage indexes.
            // TODO:  PostgreSQL dialect specific;  make more generic
            /*
            Connection connection = session.connection();
            Statement statement = connection.createStatement();

            try
            {
               String table = loader.getTable();
               if (log.isInfoEnabled())
               {
                  log.info("Analyzing " + table);
               }
               statement.execute("analyze " + table);
               connection.commit();
            }
            catch (SQLException exc)
            {
               connection.rollback();
               Throwable next = exc.getNextException();
               if (next != null)
               {
                  log.error("Next exception:", next);
               }
               
               throw exc;
            }
            finally
            {
               statement.close();
            }
            */
         }
         catch (Exception exc)
         {
            int processed = adjustedCounter(counter, badIndex, dropCount);
            int lost = counter % batchSize;
            log.error(
               "Error processing import data from " + dataFile + ";  " +
               processed + " of ? record(s) successfully processed;  " +
               lost + " record(s) uncommitted due to this error;  " +
               dropCount + " record(s) dropped",
               exc);
            
            synchronized (complete)
            {
               fatal.add(loader.getDMOName());
            }
         }
         finally
         {
            if (session != null && session.isOpen())
            {
               session.close();
            }
            if (stream != null)
            {
               stream.close();
            }
            idContextStack.push(idCtx);
         }
         
         // Report number of records successfully imported.
         return adjustedCounter(counter, badIndex, dropCount);
      }
      
      /**
       * Generate the DDL to drop and recreate the list of indexes for the given
       * table.  If the table does not exist in the database, simply return.
       * Indexes on <code>table</code> which are not within the list are
       * ignored. All output is written to the given filename.
       * <p>
       * Indexes to be dropped are identified by the names provided in the
       * list of index definitions;  the remainder of the definition is not
       * used to identify indexes to be dropped.
       *
       * @param   dialectName
       *          Name of the Hibernate dialect to use for DDL generation.
       * @param   filename
       *          File to which the DDL should be appended.
       * @param   delimiter
       *          End of statement delimiter.
       * @param   table
       *          Table which is to be reindexed.
       * @param   indexes
       *          List of index definitions.
       * @param   last
       *          <code>true</code> if this is the last call to this method
       *          for a given run.
       * @param   highID
       *          Initial value for the identity sequence.
       */
      public void generateIndexDDL(String  dialectName,
                                   String  filename,
                                   String  delimiter,
                                   String  table,
                                   List    indexes,
                                   boolean last,
                                   long    highID)
      throws java.io.IOException,
             java.io.UnsupportedEncodingException
      {
         // possibly save off dialect for later use (WARNING: NOT thread-safe)
         if (dialect == null)
         {
            Properties props = new Properties();
            props.put(Environment.DIALECT, dialectName);
            dialect = (P2JDialect) Dialect.getDialect(props);
         }

         // setup our data structures
         List<P2JIndex> idxList = (List<P2JIndex>) indexes;
         List<P2JIndex> minimalUnique =
            P2JIndex.normalizeUniqueIndexes(idxList);
         List<P2JIndex> nonRedundant = P2JIndex.nonRedundantIndexes(idxList);
         Set<P2JIndex> remaining = new LinkedHashSet<P2JIndex>(nonRedundant);
         remaining.removeAll(minimalUnique);
         
         // open output file
         OutputStream os = openStream(filename);
         
         try
         {
            generateDropIndexes(os, delimiter, nonRedundant);
            generateCreateIndexes(os, delimiter, minimalUnique, true);
            generateCreateIndexes(os, delimiter, remaining, false);
            
            if (last)
            {
               generateIdentitySequence(os, delimiter, dialect, highID);
            }
         }
         
         finally
         {
            try
            {
               os.close();
            }
            
            catch (Exception exc)
            {
               // ignore
            }
         }
      }
      
      /**
       * Drop and recreate the list of indexes for the given table.  If the
       * table does not exist in the database, simply return.  Indexes on
       * <code>table</code> which are not within the list are ignored.
       * <p>
       * Indexes to be dropped are identified by the names provided in the
       * list of index definitions;  the remainder of the definition is not
       * used to identify indexes to be dropped.
       *
       * @param   table
       *          Table which is to be reindexed.
       * @param   indexes
       *          List of index definitions.
       *
       * @throws  SQLException
       *          if any error occurs opening or closing JDBC statements.
       * @throws  HibernateException
       *          if any error occurs opening or closing sessions, or
       *          obtaining database connections.
       */
      public void reindex(String table, List indexes)
      throws SQLException
      {
         List<P2JIndex> idxList = (List<P2JIndex>) indexes;
         List<P2JIndex> minimalUnique =
            P2JIndex.normalizeUniqueIndexes(idxList);
         List<P2JIndex> nonRedundant = P2JIndex.nonRedundantIndexes(idxList);
         Set<P2JIndex> remaining = new LinkedHashSet<P2JIndex>(nonRedundant);
         remaining.removeAll(minimalUnique);
         Session session = null;
         
         try
         {
            dropIndexes(nonRedundant);
            session = openSession();
            createIndexes(session, minimalUnique, true);
            createIndexes(session, remaining, false);
         }
         finally
         {
            if (session != null)
            {
               session.close();
            }
         }
      }
      
      /**
       * Return the next import bundle which is eligible for import.  A
       * bundle is eligible for import if all of the tables upon which it is
       * dependent already have been loaded.  Bundles are returned from this
       * method in the order in which they were ranked, except that a bundle
       * with a lower rank may be returned if a higher ranking bundle still
       * has unresolved dependencies.  If there are no more bundles in the
       * job queue, the import run is finished, and this method will return
       * <code>null</code>.
       * <p>
       * This method will block if the maximum number of active threads has
       * been reached.  It will unblock once a slot becomes available.
       *
       * @return  The next import bundle to be processed, or <code>null</code>
       *          if no jobs remain in the queue.
       *
       * @throws  InterruptedException
       *          if the running thread is interrupted while waiting for an
       *          import thread slot to become available.
       */
      private ImportBundle nextJob()
      throws InterruptedException
      {
         // Keep trying to find a job as long as the queue is not empty.
         while (!jobQueue.isEmpty())
         {
            synchronized (complete)
            {
               // Wait for an available thread slot.
               while (activeThreads == maximumThreads)
               {
                  complete.wait();
               }
               
               // Iterate through the job queue until we find a job that is
               // ready to go (has all of its dependencies resolved).
               Iterator<ImportBundle> iter = jobQueue.iterator();
               while (iter.hasNext())
               {
                  ImportBundle bundle = iter.next();
                  Set<String> dependencies =
                     bundle.getLoader().getDependencies();
                  if (complete.containsAll(dependencies))
                  {
                     // We found one.  Remove it from the queue and return
                     // it.
                     iter.remove();
                     
                     return bundle;
                  }
               }
               
               // We cycled through the whole job queue, but nothing was
               // ready;  wait for a job to complete before going around
               // again.
               complete.wait();
            }
         }
         
         return null;
      }
      
      /**
       * Import a single table on a dedicated import thread.  This method
       * will block until a thread slot becomes available, if the maximum
       * number of simultaneous threads already is running.
       * <p>
       * Once a thread slot opens up, a new thread is launched to import the
       * table associated with <code>bundle</code>.  If this table is
       * dependent upon the import of other tables, and not all of these
       * tables have yet been imported, the new thread will block until such
       * time as all such dependencies have been resolved.  If any tables
       * upon which the target table is dependent have encountered fatal
       * errors, the import of the target table is aborted and the target
       * table is added to the set of tables with fatal errors.
       *
       * @param   bundle
       *          Import bundle which holds all information necessar to
       *          import data into a single table.
       * @param   dateOrder
       *          Date order string read from configuration, or null if there
       *          was no <code>date-order</code> configured. This must be set
       *          on the job thread because it is managed as a context local
       *          setting.
       *
       * @see     #setMaximumThreads
       */
      private void importAsync(final ImportBundle bundle,
                               final String dateOrder)
      {
         Runnable runner = new Runnable()
         {
            public void run()
            {
               if (dateOrder != null)
               {
                  date.setDateOrder(dateOrder);
               }
               
               SqlRecordLoader loader = bundle.getLoader();
               String table = loader.getTable();
               String dmoName = loader.getDMOName();
               Set<String> dependencies = loader.getDependencies();
               Set<String> broken = new HashSet<String>(dependencies);
               
               // Determine the intersection of this table's dependencies and
               // tables which encountered fatal errors.
               synchronized (complete)
               {
                  broken.retainAll(fatal);
               }
               
               StringBuilder buf = new StringBuilder("BEGIN ");
               buf.append(table);
               buf.append(" (");
               buf.append(loader.getDataFile().getName());
               buf.append(")");
               
               System.out.println(buf.toString());
               
               int count = 0;
               long bytes = 0L;
               boolean failed = !broken.isEmpty();
               if (failed)
               {
                  // One or more of this table's dependencies encountered a
                  // fatal error, so this table may not be imported safely.
                  // Abort this job.
                  System.err.println("Unable to import table " +
                                     table +
                                     " due to broken dependencies:  " +
                                     broken);
               }
               else
               {
                  // All dependencies have been resolved.  Run the import
                  // job.
                  count = importTable(loader, bundle.getIndexes());
                  bytes = bundle.getFileLength();
               }
               
               synchronized (complete)
               {
                  if (failed)
                  {
                     fatal.add(dmoName);
                  }
                  
                  synchronized (countLock)
                  {
                     imported += count;
                     importedBytes += bytes;
                     
                     buf = new StringBuilder("END ");
                     buf.append(table);
                     buf.append(";  ");
                     buf.append(count);
                     buf.append(" records imported");
                     messageWithStats(buf.toString());
                  }
                  
                  // Update the set of completed jobs, regardless of success.
                  complete.add(dmoName);
                  
                  // Decrement the number of active threads and notify any
                  // waiting threads.
                  activeThreads--;
                  complete.notifyAll();
               }
            }
         };
         
         synchronized (complete)
         {
            // Update the set of attempted import jobs.
            attempted.add(bundle.getLoader().getDMOName());
            
            // Increment the number of active threads.
            activeThreads++;
         }
         
         // Start the import job thread.
         Thread t = new Thread(runner);
         t.start();
      }
      
      /**
       * Print a message to stdout, followed by the current import statistics.
       * Statistics consist of the number of records imported since the start
       * of the import, and average counts of records imported per second,
       * per minute, and per hour.  Statistics are printed after the core
       * message, on a separate line.
       *
       * @param   message
       *          The core message to be printed.  May be <code>null</code>
       *          if only statistics are desired.
       */
      private void messageWithStats(String message)
      {
         if (message != null)
         {
            System.out.println(message);
         }
         
         synchronized (countLock)
         {
            long elapsed = System.currentTimeMillis() - startTime;
            if (elapsed == 0)
            {
               System.out.println("0 recs in 0:00:00.000");
               
               return;
            }
            
            Object[] args = new Object[]
            {
               new Double(importedBytes / totalBytes * 100),
               new Long(imported),
               new Long(elapsed / 3600000),
               new Long(elapsed / 60000 % 60),
               new Double(elapsed / 1000.0 % 60),
               new Long((long) (imported / (elapsed / 1000.0))),
               new Long((long) (imported / (elapsed / 60000.0))),
               new Long((long) (imported / (elapsed / 3600000.0))),
            };
            System.out.println(String.format(STATS_FMT, args));
         }
      }
      
      /**
       * Sequentially create the DDL to drop all of the specified indexes from
       * the database, write this SQL to the given stream.
       *
       * @param    os
       *           Output destination.
       * @param    delimiter
       *           End of statement delimiter.
       * @param    indexes
       *           Collection of indexes to be dropped.
       */
      private void generateDropIndexes(OutputStream         os,
                                       String               delimiter,
                                       Collection<P2JIndex> indexes)
      throws java.io.IOException,
             java.io.UnsupportedEncodingException
      {
         Iterator<P2JIndex> iter = indexes.iterator();
         
         while (iter.hasNext())
         {
            P2JIndex index = iter.next();
            writeStmt(os, dialect.getDropIndexString(index), delimiter);
         }
      }
      
      /**
       * Sequentially drop all of the specified indexes from the database.
       * JDBC errors encountered executing the drop are logged but are
       * otherwise ignored, since it may very well be that the indexes to be
       * dropped have never been created (as is the case with a new import).
       *
       * @param   indexes
       *          Collection of indexes to be dropped.
       *
       * @throws  SQLException
       *          if any error occurs creating or closing a statement.
       * @throws  HibernateException
       *          if any error occurs opening or closing a session, or
       *          obtaining a database connection.
       */
      private void dropIndexes(Collection<P2JIndex> indexes)
      throws SQLException
      {
         Iterator<P2JIndex> iter = indexes.iterator();
         Session session = null;
         try
         {
            while (iter.hasNext())
            {
               P2JIndex index = iter.next();
               
               // do not attempt to drop a word index here;  it was never
               // created
               if (index.isWord())
               {
                  if (log.isInfoEnabled())
                  {
                     log.info("Skipping dropping of word index: " +
                              index.getName());
                  }
                  
                  continue;
               }
               
               if (session == null)
                  session = openSession();

               Transaction tr = null;

               try
               {
                  tr = session.beginTransaction();
                  dropIndex(session, index);
                  tr.commit();
               }
               catch (HibernateException exc)
               {
                  if (tr != null)
                     tr.rollback();
                  session.close();
                  session = null;

                  log.error(exc);
               }
            }
         }
         finally
         {
            if (session != null)
            {
               session.close();
            }
         }
      }
      
      /**
       * Drop an index based upon the specified definition. A native SQL
       * statement is executed against the database to drop the index.
       * The statement is generated by using the database-specific dialect
       * configured for the process.
       *
       * @param   session
       *          Hibernate session to use.
       * @param   index
       *          Definition of the index to be dropped.
       *
       * @throws  HibernateException
       *          if an error occurs at the database during execution of the
       *          index drop statement.
       */
      private void dropIndex(Session session, P2JIndex index)
      throws HibernateException
      {
         String dropString = dialect.getDropIndexString(index);
         
         if (log.isInfoEnabled())
         {
            log.info("SQL:  " + dropString);
         }
         session.createSQLQuery(dropString).executeUpdate();
      }
      
      /**
       * Sequentially create the DDL to create all of the specified indexes from
       * the database, write this SQL to the given stream.
       *
       * @param    os
       *           Output destination.
       * @param    delimiter
       *           End of statement delimiter.
       * @param    indexes
       *           Collection of indexes to be created.
       */
      private void generateCreateIndexes(OutputStream         os,
                                         String               delimiter,
                                         Collection<P2JIndex> indexes,
                                         boolean              unique)
      throws java.io.IOException,
             java.io.UnsupportedEncodingException
      {
         Iterator<P2JIndex> iter = indexes.iterator();
         
         while (iter.hasNext())
         {
            P2JIndex index = iter.next();
            
            String create = dialect.getCreateIndexString(index, unique, true);

            writeStmt(os, create, delimiter);
         }
      }
      
      /**
       * Create all indexes in the specified list which match the uniqueness
       * specified by <code>unique</code>. If none match, no indexes are
       * created.
       *
       * @param   session
       *          Database session in which the indexes are create.
       * @param   indexes
       *          A list of index definitions.
       * @param   unique
       *          If <code>true</code>, all unique indexes in the list are
       *          created; if <code>false</code>, all non-unique indexes in
       *          the list are created.
       *
       * @throws  SQLException
       *          if an error occurs at the database during index creation.
       */
      private void createIndexes(Session session,
                                 Collection<P2JIndex> indexes,
                                 boolean unique)
      throws SQLException
      {
         Iterator<P2JIndex> iter = indexes.iterator();
         while (iter.hasNext())
         {
            P2JIndex index = iter.next();
            
            // do not create a word index here
            // TODO:  word indexes must be handled separately
            if (index.isWord())
            {
               if (log.isInfoEnabled())
               {
                  log.info("Skipping creation of word index: " +
                           index.getName());
               }
               
               continue;
            }

            Transaction tr = null;

            try
            {
               if (!session.getTransaction().isActive())
               {
                  // start a new transaction only if there is no active one
                  tr = session.beginTransaction();
               }

               createIndex(session, index, unique);

               if (tr != null)
                  tr.commit();
            }
            catch (HibernateException exc)
            {
               if (tr != null)
                  tr.rollback();
               log.error(exc);

               throw exc;
            }
         }
      }
      
      /**
       * Create an index based upon the specified definition. A native SQL
       * statement is executed against the database to create the index.
       * The statement is generated by using database-specific dialect
       * configured for the import.
       *
       * @param   session
       *          Hibernate session to use.
       * @param   index
       *          Definition of the index to be created.
       * @param   unique
       *          If <code>true</code>, create the index as a unique index,
       *          <code>false</code> to omit the unique constraint.
       *
       * @throws  HibernateException
       *          if an error occurs at the database during index creation.
       */
      private void createIndex(Session session,
                               P2JIndex  index,
                               boolean   unique)
      throws HibernateException
      {
         // TODO:  check if index already exists before executing create
         // statement. We can use DatabaseMetadata for this, and it would
         // avoid hassles in the event of a restarted import.
         // ...
         String createString = 
            dialect.getCreateIndexString(index, unique, false);
         
         if (log.isInfoEnabled())
         {
            log.info("SQL:  " +
                     createString.replaceAll("\t\n\r", "\\\\t\\\\n\\\\r"));
         }
         session.createSQLQuery(createString).executeUpdate();
      }
      
      /**
       * Adjust the given counter to take the specified <code>badIndex</code>
       * and <code>dropCount</code> values into account.
       *
       * @param   counter
       *          Number of total records read, but not necessarily imported.
       * @param   badIndex
       *          Index of the first problem record in an import batch, or
       *          -1 if there was no problem record. The index is zero-based,
       *          relative to the beginning of the batch.
       * @param   dropCount
       *          Cumulative number of problem records which have been
       *          dropped from the import of the current table.
       *
       * @return  Number of records successfully imported for the current
       *          tabel.
       */
      private int adjustedCounter(int counter, int badIndex, int dropCount)
      {
         if (badIndex >= 0)
         {
            return (counter - batchSize + badIndex);
         }
         
         return (counter - dropCount);
      }
      
      /**
       * Remove a problem record from a batch of records which failed on
       * import, if it can be determined which record caused the problem.
       * This is determined by walking the root cause exceptions chained to
       * the exception <code>exc</code> until an exception of type
       * <code>BatchUpdateException</code> is found. If such a cause is
       * found, the number of successful updates is extracted. This number
       * represents the index of the first record which was not successful.
       * It is this record which is removed from the batch.
       * <p>
       * An error message is logged recording the 1-based index of the
       * problem record, the data file from which it was read, and the
       * nature of the error as extracted from <code>exc</code>.
       * <p>
       * The purpose of this method is to preserve information regarding the
       * failure, while dropping the problem record in order to clean up the
       * current batch of records to enable a retry of the batch insert.
       *
       * @param   records
       *          List of records which comprised the failed batch.
       * @param   exc
       *          Exception which was caught on batch insert failure.
       * @param   counter
       *          Number of total records read up to this point. This number
       *          will be at most <code>batchSize</code> higher than the
       *          number of records successfully imported to this point.
       * @param   dataFile
       *          Data file from which data is being read, for error logging
       *          purposes.
       *
       * @return  The zero-based index, relative to the beginning of
       *          <code>records</code>, of the problem record.
       *
       * @throws  HibernateException
       *          in the event <code>exc</code> does not contain a chained
       *          <code>BatchUpdateException</code>. In this case, the index
       *          of a problem record cannot be determined, and it is likely
       *          that the error is more severe than a single bad record.
       */
      private int removeFailingRecord(List<Object> records,
                                      HibernateException exc,
                                      int counter,
                                      File dataFile)
      throws HibernateException
      {
         // Try to isolate and remove offending record.
         BatchUpdateException bue = null;
         Throwable next = exc;
         while (next != null)
         {
            if (next instanceof BatchUpdateException)
            {
               bue = (BatchUpdateException) next;
               break;
            }
            next = next.getCause();
         }
         
         // Unexpected problem; bail.
         if (bue == null)
         {
            throw exc;
         }
         
         // Remove offending record, dump stack trace and continue
         // processing.
         int badIndex = bue.getUpdateCounts().length;
         records.remove(badIndex);
         Throwable nextErr = bue.getNextException();
         if (nextErr == null)
         {
            nextErr = bue;
         }
         log.error(
            "Dropped record #" +
            (counter - Math.min(counter, batchSize) + badIndex + 1) +
            " in " + dataFile.getName() + " due to error:  " +
            nextErr.getMessage());
         
         return badIndex;
      }
   }
   
   /**
    * Helper class which manages all aspects of the mapping between a
    * Progress data export file and the DMO which corresponds with a
    * particular database table. The information managed by this class
    * includes:
    * <ul>
    * <li>the DMO class to be instantiated for each imported record;
    * <li>the input data file from which records are read;
    * <li>whether the export file reading process should assert the type
    *     of each data component it reads for a record (for debugging);
    * <li>a list of <code>PropertyMapper</code> objects, in the order the
    *     mapped data is expected to be encountered in the export file for
    *     each record (one per column, or one per subscript value for extent
    *     columns);
    * <li>a map of <code>QueryHelper</code> objects used to resolve foreign
    *     key associations, indexed by the name of the "foreign" class;
    * <li>a map of <code>Method</code> objects used to set a "foreign" class
    *     as property of the data model object currently being imported.
    * </ul>
    * <p>
    * An instance of this class is created for each table which needs to be
    * imported. The above properties are set as necessary, using the various
    * set, put, and add methods. Once the instance is fully configured,
    * {@link #nextRecord} is called repeatedly, until there are no more
    * records to process in the input data file.
    */
   public final static class SqlRecordLoader
   extends RecordLoader
   {
      /** Query helpers used to resolve foreign key associations */
      private final Map<String, QueryHelper> queries =
         new HashMap<String, QueryHelper>();
      
      /** Reusable map of parameter data for HQL queries */
      private final Map<String, Map<String, Object>> fkHqlMap =
         new HashMap<String, Map<String, Object>>();
      
      /** Map of setter methods used to set foreign key association DMOs */
      private final Map<String, Method> fkSetters =
         new HashMap<String, Method>();
      
      /** Name of database table */
      private String table = null;
      
      /** Map of not-null properties to their default values. */
      private Map<String, BaseDataType> notNullProps = 
         new HashMap<String, BaseDataType>();
      
      /**
       * Constructor.
       * <p>
       * Note:  the <code>checkTypes</code> flag should be set to
       * <code>false</code> for normal operations. It is expensive and should
       * be enabled for debug purposes only.
       *
       * @param   dmoClass
       *          Class which is instantiated for each record imported.
       * @param   table
       *          Name of the table associated with this loader instance.
       * @param   dataFile
       *          Data input file.
       * @param   checkTypes
       *          <code>true</code> to assert that the data type of each
       *          field read matches the expected type.
       */
      public SqlRecordLoader(Class<?> dmoClass,
                             String table,
                             File dataFile,
                             boolean checkTypes)
      {
         super(dmoClass, dataFile, checkTypes);
         
         this.table = table;
      }
      
      /**
       * Add the not-null property with its default value, if any.
       * 
       * @param   prop
       *          The property name.
       * @param   def
       *          The default value.
       */
      public void addNotNullProperty(String prop, BaseDataType def)
      {
         notNullProps.put(prop, def);
      }
      
      /**
       * Get the name of the database table associated with this loader
       * instance.
       *
       * @return  Table name.
       */
      public String getTable()
      {
         return table;
      }
      
      /**
       * Get the query helper which corresponds with a class on the opposite
       * end of a foreign association. The query helper returned uses data
       * in the current record to compose a query which should retrieve the
       * foreign object.
       *
       * @param   joinClass
       *          Name of class on the other end of a foreign association
       *          with the DMO being loaded by this object.
       *
       * @return  Query helper corresponding with <code>joinClass</code>.
       */
      public QueryHelper getQueryHelper(String joinClass)
      {
         return queries.get(joinClass);
      }
      
      /**
       * Register a query helper to a class name. This will be used to
       * compose a query which will retrieve an object which will satisfy
       * a foreign key requirement in the record currently being loaded.
       *
       * @param   joinClass
       *          Name of the class of the object which will be retrieved
       *          by the query.
       * @param   helper
       *          Helper object which will compose the query.
       */
      public void putQueryHelper(String joinClass, QueryHelper helper)
      {
         queries.put(joinClass, helper);
      }
      
      /**
       * Register a method to a class name. This method will be called via
       * reflection to store an object in the DMO currently being loaded.
       * The object stored will be retrieved via a query and will satisfy
       * a foreign key requirement in the record currently being loaded.
       *
       * @param   joinClass
       *          Name of the class of the object which represents the
       *          foreign end of a foreign key association.
       * @param   setter
       *          Method which will be invoked to store the foreign object
       *          in the DMO currently being loaded.
       */
      public void putForeignKeySetter(String joinClass, Method setter)
      {
         fkSetters.put(joinClass, setter);
      }
      
      /**
       * Process the data value most recently read from the data export file.
       * This method is called by the superclass' <code>nextRecord</code>
       * method for each data value read from the data export file. If the
       * current property is not delegated and not dropped, 
       *
       * @param   mapper
       *          Mapping object associated with the current property.
       * @param   record
       *          Java-bean-like object which holds the contents of the record
       *          currently being loaded.
       * @param   value
       *          Data value read from the data export file for the current
       *          property. It has <i>not</i> yet been stored in the
       *          <code>record</code> object.
       *
       * @throws  IllegalAccessException
       *          if the setter method which we invoke is not accessible.
       * @throws  InvocationTargetException
       *          if there is any error invoking the appropriate setter method.
       */
      protected void processNextValue(PropertyMapper mapper,
                                      Object record,
                                      Object value)
      throws IllegalAccessException,
             InvocationTargetException
      {
         // If this property has not been delegated to another class, we
         // must set the value of the data in the DMO.
         if (!mapper.isDelegated())
         {
            // Invoke the appropriate DMO setter method to store the
            // value.
            mapper.invokeSetter(record, value);
         }
         
         // Store off the value in maps of HQL parameters for foreign key
         // association resolution queries later.
         Iterator<String> iter = mapper.joins();
         while (iter.hasNext())
         {
            String joinClass = iter.next();
            Map<String, Object> hqlParms = fkHqlMap.get(joinClass);
            if (hqlParms == null)
            {
               hqlParms = new HashMap<String, Object>();
               fkHqlMap.put(joinClass, hqlParms);
            }
            hqlParms.put(mapper.getPropertyName(), value);
         }
      }
      
      /**
       * Return the set of all DMO classes which represent the tables which
       * must be loaded before this loader may import records into its
       * associated table.
       *
       * @return  Set of dependencies for this loader, or an empty set if
       *          this loader has no dependencies.
       */
      Set<String> getDependencies()
      {
         return queries.keySet();
      }
      
      /**
       * Read a single record from the current location in a Progress export
       * file and store its data in a bean-like object. Each entry read from
       * the export file represents either a field in the current table, or a
       * single value within an extent field. For each record, a new instance
       * of the record class associated with this loader is created. The order
       * in which the data is tokenized corresponds directly with the order
       * of the <code>PropertyMapper</code> objects stored in this loader
       * instance.
       * <p>
       * <b>The <code>checkTypes</code> functionality is not honored here
       * although this would be the correct location to add that feature.</b>
       * <p>
       * This method must consider columns which have been delegated to other
       * tables, as well as foreign key associations to other tables. If a
       * column is encountered which is flagged as a delegated column, its
       * value is read and stored, but is not set into the DMO (no setter
       * method will be available in this case).
       * <p>
       * If one or more foreign key associations must be resolved for the
       * current record, a predefined HQL query is executed against Hibernate
       * for each such association, using a map of data values read for the
       * current record to bind as parameters.
       *
       * @param   input
       *          The export file to be read.
       * @param   session
       *          Hibernate session used to execute any queries necessary to
       *          resolve foreign key associations.
       *
       * @return  A data model object instance containing the data for the
       *          next record in the export file, or <code>null</code> if no
       *          more records are available.
       *
       * @throws  InstantiationException
       *          if any error occurs creating a new instance of the DMO
       *          class.
       * @throws  IllegalAccessException
       *          if the DMO class or its default constructor is not
       *          accessible.
       * @throws  InvocationTargetException
       *          if there is any error invoking a setter method for an
       *          individual record field.
       *
       * @see     #findUnique
       */
      Object nextRecord(Stream input, Session session)
      throws InstantiationException,
             IllegalAccessException,
             InvocationTargetException
      {
         fkHqlMap.clear();
         Object dmo = nextRecord(input);
         
         if (dmo != null)
         {
            // Execute queries to resolve foreign keys and set the results
            // into the DMO.
            Iterator<String> iter = queries.keySet().iterator();
            while (iter.hasNext())
            {
               String joinClass = iter.next();
               QueryHelper helper = queries.get(joinClass);
               Map<String, Object> hqlParms = fkHqlMap.get(joinClass);
               Object value = findUnique(helper, hqlParms, session);
               Object args[] = { value };
               Method setter = fkSetters.get(joinClass);
               setter.invoke(dmo, args);
            }
         }
         
         return dmo;
      }
      
      /**
       * Execute the given query to produce a unique result. This executes
       * the minimal query necessary to return the primary key of the DMO
       * which matches the query criteria. The Hibernate session is then used
       * to create and return an uninitialized proxy instance of the DMO. The
       * proxy is never actually initialized. This implementation is used to
       * avoid the overhead of:
       * <ul>
       *   <li>retrieving all data for the matching record from the database;
       *   <li>hydrating the DMO with that data.
       * </ul>
       * <p>
       * Using an uninitialized proxy is sufficient, because we do not
       * actually need the record's data. We only need its primary key, to
       * satisfy the foreign key requirement of the DMO we are currently
       * preparing for insertion into the current table.
       *
       * @param   helper
       *          Helper object which manages the Hibernate query to be
       *          executed.
       * @param   parms
       *          A map of parameter names to parameter value objects. Each
       *          entry is set into the query to replace the named parameter
       *          placeholders defined for the query.
       * @param   session
       *          Hibernate database session object used to create the
       *          Hibernate query to be executed.
       *
       * @return  The query's unique result (a DMO instance), or
       *          <code>null</code> if no record matched the criteria
       *          specified by the query.
       *
       * @throws  HibernateException
       *          if any problem occurred executing the query, or if the
       *          query returned a non-unique result.
       */
      private Object findUnique(QueryHelper helper,
                                Map<String, Object> parms,
                                Session session)
      throws HibernateException
      {
         Query query = helper.getQuery(session);
         Map<String, QueryHelper.NameType> nameTypeMap =
            helper.getNameTypeMap();
         
         Iterator<String> iter = parms.keySet().iterator();
         while (iter.hasNext())
         {
            String localName = iter.next();
            Object parm = parms.get(localName);
            QueryHelper.NameType nameType = nameTypeMap.get(localName);
            query.setParameter(nameType.name, parm, nameType.type);
         }
         
         Serializable id = (Serializable) query.uniqueResult();
         if (id == null)
         {
            return null;
         }
         
         return session.load(helper.getDMOClass(), id);
      }
   }
   
   /**
    * A helper class to manage the information necessary to compose and
    * execute a database query using HQL (Hibernate Query Language). This
    * is used during data import to perform queries when resolving foreign
    * key associations.
    * <p>
    * An instance of this class is constructed for a particular, target DMO.
    * Any number of typed, named parameters may be added using the {@link
    * #addParameter} method. Once all named parameters have been added,
    * {@link #getQuery} is invoked to build and/or retrieve the resulting
    * Hibernate query. The underlying HQL query string which is composed is
    * of the form:
    * <pre>
    *    select dmo.id
    *    from <dmo_name> dmo
    *    where dmo.<parm1> = :parm1
    *          [and dmo.<parm2> = :parm2
    *           ...
    *           [and dmo.<parm_n> = :<parm_n>]]
    * </pre>
    * <p>
    * Note that the result of executing the query is the primary key of the
    * DMO which matches the query's criteria (or <code>null</code> if no
    * match is found).
    */
   public static class QueryHelper
   {
      /** Mapping of data type names to Hibernate <code>Type</code>s */
      private static final Map<String, Type> TYPES =
         new HashMap<String, Type>();
      
      // Initialize the TYPES map.
      static
      {
         TYPES.put("integer"  , new CustomType(new IntegerUserType()));
         TYPES.put("decimal"  , new CustomType(new DecimalUserType()));
         TYPES.put("character", new CustomType(new CharacterUserType()));
         TYPES.put("logical"  , new CustomType(new LogicalUserType()));
         TYPES.put("date"     , new CustomType(new DateUserType()));
         TYPES.put("raw"      , new CustomType(new RawUserType()));
         TYPES.put("rowid"    , new CustomType(new RowIDUserType()));
      }
      
      /** Class of the DMO which is to be returned as the query result */
      private Class<?> dmoClass = null;
      
      /** Map of query local parameter names to <code>NameType</code>s */
      private Map<String, NameType> nameTypeMap =
         new LinkedHashMap<String, NameType>();
      
      /** Hibernate query object */
      private Query query = null;
      
      /** Cached session object */
      private Session session = null;
      
      /** Cache query results? */
      private boolean cacheable = false;
      
      /** Database dialect in use */
      private P2JDialect dialect = null;
      
      /**
       * Constructor.
       *
       * @param   dialect
       *          Database dialect in use.
       * @param   dmoClass
       *          Implementation class of the DMO which is to be returned as
       *          the query result.
       * @param   cacheable
       *          <code>true</code> to enable query results to be cached;
       *          <code>false</code> to disable this feature. This is only
       *          useful to enable for many-to-one associations, where a
       *          query is likely to be repeated.
       */
      public QueryHelper(P2JDialect dialect,
                         Class<?> dmoClass, 
                         boolean cacheable)
      {
         this.dialect = dialect;
         this.dmoClass = dmoClass;
         this.cacheable = cacheable;
      }
      
      /**
       * Add a named parameter to the query and determine its Hibernate
       * <code>Type</code>. This will resolve to an equality test in the
       * the where clause of the HQL query (as in: <code>from MyClass dmo
       * where <b>dmo.myParam = :myParam</b></code>...).
       *
       * @param   localName
       *          Name of the parameter property in the local class (i.e.,
       *          the near end of the association).
       * @param   foreignName
       *          Name of the parameter property in the foreign class (i.e.,
       *          the far end of the association). This is often the same
       *          as <code>localName</code>, but may be different if the
       *          names got out of synch during name conversion from the
       *          Progress schema to the data model representation.
       * @param   type
       *          Type name for the parameter; typically the unqualified
       *          class name or primitive name of the data type of the
       *          parameter.
       * @param   ignoreCase
       *          <code>true</code> to ignore case in the query (applicable to
       *          text properties only);  <code>false</code> for a case
       *          sensitive query.
       *
       * @throws  IllegalStateException
       *          if this method is called after the {@link #getQuery}
       *          method is invoked for a particular helper instance. This
       *          is because <code>getQuery</code> assumes that no more
       *          parameters will be added once a query has been built and
       *          cached.
       */
      public void addParameter(String localName,
                               String foreignName,
                               String type,
                               boolean ignoreCase)
      {
         if (query != null)
         {
            throw new IllegalStateException(
              "Cannot add property;  query already has been created:  " +
              query);
         }
         
         nameTypeMap.put(localName,
                         new NameType(foreignName,
                                      TYPES.get(type),
                                      ignoreCase));
      }
      
      /**
       * Get a mapping of all local parameter names parameters for this query
       * to their foreign property names and Hibernate <code>Type</code>s.
       *
       * @return  An unmodifiable map as described above.
       */
      public Map<String, NameType> getNameTypeMap()
      {
         return Collections.unmodifiableMap(nameTypeMap);
      }
      
      /**
       * Get the Hibernate query object generated by this query helper,
       * based upon the DMO name supplied at construction and all the named
       * parameters supplied via the {@link #addParameter} method. Once this
       * method is called for a particular helper instance, no more named
       * parameters may be added, since this method builds and caches the
       * query the first time it is executed. Thereafter, it simply returns
       * the cached query, as long as <code>session</code> matches the
       * session in use at the time the cached query was created. If
       * <code>session</code> is new, the query is rebuilt and recached.
       *
       * @param   session
       *          Hibernate session object used to create the query.
       *
       * @return  Hibernate query.
       */
      public Query getQuery(Session session)
      {
         if (query != null && this.session.equals(session))
         {
            return query;
         }
         
         this.session = session;
         
         StringBuilder buf = new StringBuilder("select dmo.id from ");
         String name = dmoClass.getName();
         buf.append(name.substring(name.lastIndexOf(".") + 1));
         
         Type charType = TYPES.get("character");
         
         Iterator<String> iter = nameTypeMap.keySet().iterator();
         for (int i = 0; iter.hasNext(); i++)
         {
            if (i == 0)
            {
               buf.append(" dmo where ");
            }
            else
            {
               buf.append(" and ");
            }
            
            NameType nameType = nameTypeMap.get(iter.next());
            
            boolean trim = (nameType.type == charType);
            
            StringBuilder sb = new StringBuilder("dmo.");
            sb.append(nameType.name);
            DBUtils.composePropertyName(trim,
                                        nameType.ignoreCase,
                                        dialect.getComputedColumnPrefix(),
                                        null,
                                        sb.toString(),
                                        buf);
            
            buf.append(" = ");
            
            sb.setLength(0);
            sb.append(":");
            sb.append(nameType.name);
            DBUtils.composePropertyName(trim,
                                        nameType.ignoreCase,
                                        dialect.getComputedColumnPrefix(),
                                        null,
                                        sb.toString(),
                                        buf);
         }
         
         query = session.createQuery(buf.toString());
         query.setCacheable(cacheable);
         query.setReadOnly(true);
         
         if (log.isInfoEnabled())
         {
            log.info("Creating query:  [" + query.getQueryString() + "]");
         }
         
         return query;
      }
      
      /**
       * Get the class of the DMO implementation instance which is to be
       * returned from executing this helper's query.
       *
       * @return  DMO implementation class.
       */
      Class<?> getDMOClass()
      {
         return dmoClass;
      }
      
      /**
       * Simple container which holds a parameter's property name and its
       * Hibernate type, for purposes of substituting parameter values into
       * queries at runtime. The name is that of the parameter's property
       * in the foreign class, which may differ from the joining property in
       * the local class, if the name conversion step of schema conversion
       * changed one of these names and not the other (to avoid a naming
       * conflict, for instance).
       * <p>
       * Instance variables are accessed directly, rather than through
       * getters, to avoid the additional overhead of a method call. Objects
       * of this class are immutable.
       */
      static class NameType
      {
         /** Name of parameter property in foreign class */
         final String name;
         
         /** Hibernate type of parameter */
         final Type type;
         
         /** If case must be ignored (only for a character parameter) */
         final boolean ignoreCase;
         
         /**
          * Constructor.
          *
          * @param   name
          *          Name of parameter in foreign class.
          * @param   type
          *          Hibernate parameter type.
          * @param   ignoreCase
          *          <code>true</code> to ignore case;  else
          *          <code>false</code>.
          */
         private NameType(String name, Type type, boolean ignoreCase)
         {
            this.name = name;
            this.type = type;
            this.ignoreCase = ignoreCase;
         }
      }
   }
   
   /**
    * All information necessary to perform the import of a single table, and
    * to prioritize that import among all tables to be imported in the same
    * run.
    */
   private class ImportBundle
   implements Comparable<ImportBundle>
   {
      /** Record loader object */
      private final SqlRecordLoader loader;
      
      /** List of all indexes to be created for the target table */
      private final List<P2JIndex> indexes;
      
      /** Length in bytes of the exported data file */
      private final long fileLength;
      
      /** Import bundles dependent upon this one to finish its import */
      private final Set<ImportBundle> dependents =
         new HashSet<ImportBundle>();
      
      /** Load time heuristic */
      private double weight = 0.0;
      
      /** Import order sorting heuristic */
      private double rank = -1.0;
      
      /**
       * Constructor.
       *
       * @param   loader
       *          Record loader object.
       * @param   indexes
       *          List of all indexes to be created for the target table.
       */
      ImportBundle(SqlRecordLoader loader, List<P2JIndex> indexes)
      {
         this.loader = loader;
         this.indexes = indexes;
         File file = loader.getDataFile();
         this.fileLength = (file.exists() ? file.length() : 0L);
      }
      
      /**
       * Compare this object with another of the same type for sorting
       * purposes.  Import bundles with the highest calculated rank heuristic
       * are sorted first, so that they will be scheduled for import earlier
       * in the import job.
       *
       * @param   that
       *          <code>ImportBundle</code> with which to compare this
       *          instance for sorting purposes.
       *
       * @return  Positive number to sort this instance before the given
       *          object;  negative number to sort this instance after the
       *          given object;  0 to give both objects equal weight.
       */
      public int compareTo(ImportBundle that)
      {
         String thisName = this.loader.getDMOName();
         String thatName = that.loader.getDMOName();
         
         if (this.rank > that.rank)
         {
            return -1;
         }
         
         if (that.rank > this.rank)
         {
            return 1;
         }
         
         return thisName.compareTo(thatName);
      }
      
      /**
       * Produce a string representation of this object, primarily for debug
       * purposes.
       *
       * @return  String which describes this object.
       */
      public String toString()
      {
         StringBuilder buf = new StringBuilder(loader.getDMOName());
         
         buf.append(":");
         buf.append(rank);
         buf.append(" [");
         buf.append(fileLength);
         buf.append(" bytes]");
         
         Set<String> deps = loader.getDependencies();
         if (!deps.isEmpty())
         {
            buf.append("\n   needs ");
            buf.append(new TreeSet<String>(deps));
         }
         
         Iterator<ImportBundle> iter = dependents.iterator();
         for (int i = 0; iter.hasNext(); i++)
         {
            ImportBundle next = iter.next();
            if (i == 0)
            {
               buf.append("\n   needed by [");
            }
            buf.append(next.getLoader().getDMOName());
            buf.append(iter.hasNext() ? ", " : "]");
         }
         
         return buf.toString();
      }
      
      /**
       * Get the record loader associated with this bundle.
       *
       * @return  Record loader.
       */
      SqlRecordLoader getLoader()
      {
         return loader;
      }
      
      /**
       * Get the indexes to be created for the target table.
       *
       * @return  Indexes.
       */
      List<P2JIndex> getIndexes()
      {
         return indexes;
      }
      
      /**
       * Add to this bundle's set of import bundles which are dependent upon
       * it to load first.  This relationship is used to determine the import
       * order across all tables in the import job.
       *
       * @param   bundle
       *          An import bundle for a table which is dependent upon this
       *          bundle's table loading first.
       */
      void addDependent(ImportBundle bundle)
      {
         dependents.add(bundle);
      }
      
      /**
       * Get this import bundle's weight, which is a heuristic intended to
       * measure the relative time it will take to load this table.
       *
       * @return  Weight heuristic.
       */
      double getWeight()
      {
         return weight;
      }
      
      /**
       * Get this import bundle's rank, which is a heuristic intended to rank
       * the importance of loading this table sooner than other tables.
       *
       * @return  Rank heuristic.
       */
      double getRank()
      {
         return rank;
      }
      
      /**
       * Get the number of bytes in the exported data file for this table.
       *
       * @return  Byte count.
       */
      long getFileLength()
      {
         return fileLength;
      }
      
      /**
       * Calculate this import bundle's weight, which is a heuristic intended
       * to measure the relative time it will take to load this table.  In
       * making this determination, the weight of each of this table's
       * dependencies is factored in, each of which may have the weight of
       * its dependencies factored in, and so on.  A bundle's weight is used
       * downstream in ranking bundles relative to one another when
       * determining import order.
       *
       * @see  #calculateRank
       */
      void calculateWeight()
      {
         weight = fileLength * 2;
         
         Iterator<String> iter = loader.getDependencies().iterator();
         while (iter.hasNext())
         {
            String key = iter.next();
            ImportBundle dependency = importBundles.get(key);
            if (dependency != null)
            {
               weight += (dependency.fileLength * 0.75);
            }
         }
      }
      
      /**
       * Calculate this import bundle's rank, which is a heuristic intended
       * to rank the importance of loading this table sooner than other
       * tables.  In making this determination, this bundle's weight is taken
       * into consideration (calculated previously), as are the ranks of all
       * those tables which are dependent upon this table to load first.
       * This algorithm ensures that a table always has a higher rank (and
       * thus is sorted to load before) any tables which are dependent upon
       * it.
       *
       * @see  #calculateWeight
       */
      void calculateRank()
      {
         if (rank < 0.0)
         {
            rank = weight;
            
            Iterator<ImportBundle> iter = dependents.iterator();
            while (iter.hasNext())
            {
               ImportBundle bundle = iter.next();
               bundle.calculateRank();
               rank += bundle.getRank();
            }
         }
      }
   }
   
   /**
    * Thread local context which tracks next available primary key ID.  IDs
    * are unique across all import threads and are batched together within
    * increments of {@link ImportWorker#idBracketSize}.
    */
   private class IDContext
   {
      /** Counter */
      private long counter;
      
      /**
       * Default constructor;  sets up internal counter.
       */
      IDContext()
      {
         prepareNextBracket();
      }
      
      /**
       * Get the next available, globally unique ID.  When next ID falls on
       * a bracket boundary, it is returned, but the next bracket of IDs for
       * this thread is prepared.
       * 
       * @return  Next, globally unique ID.
       */
      long getNextID()
      {
         long result = ++counter;
         
         if ((result % idBracketSize) == 0)
         {
            prepareNextBracket();
         }
         
         return result;
      }
      
      /**
       * Get the most recently allocated, globally unique ID.
       * 
       * @return  Current, globally unique ID.
       */
      long getCurrentID()
      {
         return counter;
      }
      
      /**
       * Prepare the next bracket of available IDs for this context.
       */
      private void prepareNextBracket()
      {
         synchronized (idContextStack)
         {
            counter = nextGlobalID;
            nextGlobalID += idBracketSize;
         }
      }
   }
}
